from __future__ import annotations

import os
import re
from pathlib import Path

from pyaerocom.units import Unit
import pytest
import xarray as xr

import pyaerocom.exceptions as exc
from pyaerocom import get_variable
from pyaerocom.griddeddata import GriddedData
from pyaerocom.io.mscw_ctm.reader import ReadEMEP, ReadMscwCtm
from tests.conftest import TEST_RTOL
from tests.fixtures.mscw_ctm import create_fake_MSCWCtm_data


@pytest.fixture()
def reader() -> ReadMscwCtm:
    """empty EMEP MSCW-CTM reader"""
    return ReadMscwCtm()


@pytest.fixture()
def data_dir(path_emep: dict[str, str]) -> str:
    """path to EMEP test data"""
    return path_emep["data_dir"]


def test_ReadMscwCtm__get_year_from_nc(data_dir: str):
    yr = ReadMscwCtm._get_year_from_nc(os.path.join(data_dir, "Base_fullrun.nc"))
    assert yr == 2017
    yr = ReadMscwCtm._get_year_from_nc(os.path.join(data_dir, "Base_day.nc"))
    assert yr == 2017
    yr = ReadMscwCtm._get_year_from_nc(os.path.join(data_dir, "Base_month.nc"))
    assert yr == 2017


def test_ReadMscwCtm__init__(data_dir: str):
    reader = ReadMscwCtm("EMEP_2017", data_dir)
    assert getattr(reader, "data_id") == "EMEP_2017"
    assert getattr(reader, "_data_dir") == data_dir


def test_ReadMscwCtm__init___error():
    data_dir = "not_a_real_path"
    with pytest.raises(FileNotFoundError) as e:
        ReadMscwCtm(None, data_dir)
    assert str(e.value) == data_dir


def test_ReadMscwCtm_data_dir(data_dir: str):
    reader = ReadMscwCtm()
    reader._data_dir = data_dir
    assert Path(reader._data_dir) == Path(data_dir)


@pytest.mark.parametrize(
    "value,exception,error",
    [
        (None, ValueError, "Data dir None needs to be a dictionary or a file"),
        ("not_a_real_path", FileNotFoundError, "not_a_real_path"),
    ],
)
def test_ReadMscwCtm_data_dir_error(value, exception, error: str):
    reader = ReadMscwCtm(value)
    with pytest.raises(exception) as e:
        reader._data_dir = value
    assert str(e.value) == error


def test__ReadMscwCtm__check_files_in_data_dir(data_dir: str):
    reader = ReadMscwCtm()
    matches = reader._check_files_in_data_dir(data_dir)
    assert len(matches) == 3


def test__ReadMscwCtm__check_files_in_data_dir_error():
    reader = ReadMscwCtm()
    with pytest.raises(FileNotFoundError):
        reader._check_files_in_data_dir("/tmp")


def test_ReadMscwCtm_ts_type():
    reader = ReadMscwCtm()
    assert reader._ts_type == "daily"


def test_ReadMscwCtm_var_map():
    var_map = ReadMscwCtm().var_map
    assert isinstance(var_map, dict)


@pytest.mark.parametrize(
    "var_name, ts_type", [("vmro3", "daily"), ("vmro3", None), ("concpmgt25", "daily")]
)
def test_ReadMscwCtm_read_var(var_name: str, ts_type: str, data_dir: str):
    reader = ReadMscwCtm(data_dir=data_dir)
    data = reader.read_var(var_name, ts_type)
    assert isinstance(data, GriddedData)
    if ts_type is not None:
        assert data.ts_type == ts_type
    assert data.ts_type is not None
    assert data.ts_type == reader._ts_type


@pytest.mark.parametrize(
    "var_name, ts_type, exception, error",
    [
        (
            "blaaa",
            "daily",
            exc.VariableDefinitionError,
            "Error (VarCollection): input variable blaaa is not supported",
        ),
        ("od550gt1aer", "daily", exc.VarNotAvailableError, "od550gt1aer"),
    ],
)
def test_ReadMscwCtm_read_var_error(
    var_name: str, ts_type: str, exception: type[Exception], error: str, data_dir: str
):
    reader = ReadMscwCtm(data_dir=data_dir)
    with pytest.raises(exception) as e:
        reader.read_var(var_name, ts_type)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "var_name, ts_type",
    [
        ("concpmgt25", "daily"),
        ("concpmgt25", "monthly"),
    ],
)
def test_ReadMscwCtm__compute_var(var_name, ts_type, data_dir: str):
    reader = ReadMscwCtm(data_dir=data_dir)
    data, proj_info = reader._compute_var(var_name, ts_type)
    assert isinstance(data, xr.DataArray)


def test_ReadMscwCtm__compute_var_error(data_dir: str):
    reader = ReadMscwCtm(data_dir=data_dir)
    with pytest.raises(KeyError):
        reader._compute_var("blaaa", "daily")


def test_ReadMscwCtm_data(data_dir: str):
    reader = ReadMscwCtm(data_dir=data_dir)

    vars_provided = reader.vars_provided
    assert isinstance(vars_provided, list)
    assert "vmro3" in vars_provided

    data = reader.read_var("vmro3", ts_type="daily")
    assert isinstance(data, GriddedData)
    assert data.time.long_name == "time"
    assert data.time.standard_name == "time"
    assert data.ts_type == "daily"

    data = reader.read_var("vmro3")
    assert isinstance(data, GriddedData)
    assert data.time.long_name == "time"
    assert data.time.standard_name == "time"
    assert data.ts_type == "daily"


def test_ReadMscwCtm_directory(data_dir: str):
    reader = ReadMscwCtm(data_dir=data_dir)
    assert reader._data_dir == data_dir
    vars_provided = reader.vars_provided
    assert "vmro3" in vars_provided
    assert "concpm10" in vars_provided
    assert "concno2" in vars_provided
    paths = reader._filepaths
    assert len(paths) == 3


@pytest.mark.parametrize(
    "filename,ts_type",
    [
        ("Base_hour.nc", "hourly"),
        ("Base_month.nc", "monthly"),
        ("Base_day.nc", "daily"),
        ("Base_fullrun", "yearly"),
    ],
)
def test_ReadMscwCtm_ts_type_from_filename(reader, filename, ts_type):
    assert reader._ts_type_from_filename(filename) == ts_type


def test_ReadMscwCtm_ts_type_from_filename_error(reader):
    with pytest.raises(ValueError) as e:
        reader._ts_type_from_filename("blaaa")
    assert str(e.value) == "Failed to retrieve ts_type from filename blaaa"


@pytest.mark.parametrize(
    "filename,ts_type",
    [
        ("Base_hour.nc", "hourly"),
        ("Base_month.nc", "monthly"),
        ("Base_day.nc", "daily"),
        ("Base_fullrun.nc", "yearly"),
    ],
)
def test_ReadMscwCtm_filename_from_ts_type(reader, filename, ts_type):
    assert reader._filename_from_ts_type(ts_type) == filename


def test_ReadMscwCtm_filename_from_ts_type_error(reader):
    with pytest.raises(ValueError) as e:
        reader._filename_from_ts_type("blaaa")
    assert str(e.value) == "unknown ts_type=blaaa"


def test_ReadMscwCtm_years_avail(data_dir: str):
    reader = ReadMscwCtm(data_dir=data_dir)
    assert reader.years_avail == ["2017"]


def test_ReadMscwCtm_preprocess_units():
    units = ""
    prefix = "AOD"
    assert ReadMscwCtm()._preprocess_units(units, prefix) == "1"


def test_ReadMscwCtm_open_file(data_dir: str):
    reader = ReadMscwCtm()
    with pytest.raises(AttributeError):
        reader._open_file()
    reader._data_dir = data_dir
    data = reader._open_file()
    assert isinstance(data, xr.Dataset)
    assert reader._filedata is data


@pytest.mark.parametrize(
    "var_name, value",
    [
        ("od550gt1aer", False),
        ("absc550aer", True),
        ("concpm10", True),
        ("sconcpm10", True),
    ],
)
def test_ReadMscwCtm_has_var(reader, var_name, value):
    assert reader.has_var(var_name) == value


def test_ReadMscwCtm_has_var_error(reader):
    with pytest.raises(exc.VariableDefinitionError, match="input variable blaa is not supported"):
        reader.has_var("blaa")


def test_ReadMscwCtm__str__():
    assert str(ReadMscwCtm()) == "ReadMscwCtm"


def test_ReadMscwCtm__repr__():
    assert repr(ReadMscwCtm()) == "ReadMscwCtm"


def test_ReadEMEP__init__():
    with pytest.warns(DeprecationWarning, match="use ReadMscwCtm instead"):
        assert isinstance(ReadEMEP(), ReadMscwCtm)


def emep_data_path(tmp_path: Path, freq: str | list[str], vars_and_units: dict[str, str]) -> Path:
    reader = ReadMscwCtm()
    varmap = reader.var_map

    root = tmp_path / "emep"
    frequencies = freq if isinstance(freq, list) else [freq]
    for freq in frequencies:
        for year in ["2017", "2018", "2019", "2015", "2018", "2013"]:
            ds = xr.Dataset()
            for var_name, units in vars_and_units.items():
                var_name = varmap[var_name]
                ds[var_name] = create_fake_MSCWCtm_data(year=year, tst=reader.FREQ_CODES[freq])
                ds[var_name].attrs.update(units=units, var_name=var_name)

            path = root / str(year) / f"Base_{freq}.nc"
            path.parent.mkdir(exist_ok=True, parents=True)
            ds.to_netcdf(path)
    return root


@pytest.fixture
def data_path(tmp_path: Path, freq: str | list[str], vars_and_units: dict[str, str]) -> Path:
    """emep test data on a temporary path"""
    return emep_data_path(tmp_path, freq, vars_and_units)


def test_ReadMscwCtm_aux_var_defs():
    req = ReadMscwCtm.AUX_REQUIRES
    funs = ReadMscwCtm.AUX_FUNS
    assert len(req) == len(funs)
    assert all([x in funs.keys() for x in req])


M_N = 14.006
M_O = 15.999
M_H = 1.007
M_HNO3 = M_H + M_N + M_O * 3
M_NO3 = M_N + M_O * 3


@pytest.mark.parametrize(
    "vars_and_units,freq,add_read,chk_mean",
    [
        ({"wetoxs": "mg S m-2"}, "day", None, {"wetoxs": 1}),
        ({"prmm": "mm"}, "hour", None, {"prmm": 24}),
        ({"prmm": "mm d-1"}, "hour", None, {"prmm": 1}),
        ({"concpm10": "ug m-3"}, "day", None, {"concpm10": 1}),
        ({"concpm10": "ug m-3"}, "hour", None, None),
        (
            {"concno3c": "ug m-3", "concno3f": "ug m-3"},
            "day",
            ["concno3"],
            {"concno3c": 1, "concno3f": 1, "concno3": 2},
        ),
        (
            {"concoxn": "ug m-3"},
            "day",
            ["concNtno3"],
            {
                "concoxn": 1,
                "concNtno3": M_N / (M_N + 3 * M_O),
            },
        ),
        ({"wetoxs": "mg S m-2 d-1"}, "day", None, {"wetoxs": 1}),
        ({"wetoxs": "Tg S m-2 d-1"}, "day", None, {"wetoxs": 1e15}),
    ],
)
def test_read_emep_dummy_data(
    data_path: Path,
    vars_and_units: dict[str, str],
    freq: str,
    add_read: list[str] | None,
    chk_mean: dict[str, float],
):
    reader = ReadMscwCtm(data_dir=str(data_path / "2017"))
    tst = reader.FREQ_CODES[freq]
    objs = {}
    for var in vars_and_units:
        data = reader.read_var(var, ts_type=tst)
        objs[var] = data
        assert isinstance(data, GriddedData)
        aerocom_unit = Unit(get_variable(var).units)
        assert Unit(data.units) == aerocom_unit
        assert data.ts_type == tst
    if isinstance(add_read, list):
        for var in add_read:
            data = reader.read_var(var, ts_type=tst)
            objs[var] = data
    if isinstance(chk_mean, dict):
        for var, mean in chk_mean.items():
            assert objs[var].cube.data.mean() == pytest.approx(mean, abs=0.1, rel=TEST_RTOL)


def test_read_emep_dummy_data_error(tmp_path: Path):
    data_path = emep_data_path(tmp_path, "day", vars_and_units={"concno3c": "ug m-3"})
    reader = ReadMscwCtm(data_dir=str(data_path / "2017"))
    with pytest.raises(exc.VarNotAvailableError) as e:
        reader.read_var("concno3", ts_type="daily")
    assert str(e.value) == "concno3f (SURF_ug_NO3_F) not available in Base_day.nc"


@pytest.mark.parametrize(
    "year,years,freq",
    [
        ("", [2013, 2015, 2017, 2018, 2019], "day"),
        ("2017", [2017], "month"),
        ("2019", [2019], "hour"),
    ],
)
@pytest.mark.parametrize("vars_and_units", [{"prmm": "mm"}])
def test_read_emep_clean_filepaths(data_path: Path, year, years: list[int], freq: str):
    reader = ReadMscwCtm(data_dir=str(data_path / year))
    tst = reader.FREQ_CODES[freq]

    filepaths = reader._filepaths

    cleaned_paths = reader._clean_filepaths(filepaths, years, tst)
    # assert len(cleaned_paths) == len(years)

    found = []
    for path in cleaned_paths:
        assert path in filepaths
        match = re.search(r".*(20\d\d).*", Path(path).parent.name)
        assert match is not None
        year = int(match.group(1))
        found.append(year)

    assert set(found) == set(years)


@pytest.mark.parametrize(
    "years,freq,error",
    [
        ([2013, 2015, 2017, 2018, 2019, 2012], "month", "A different amount of years "),
        ([2013, 2016, 2017], "day", "The year "),
    ],
)
@pytest.mark.parametrize("vars_and_units", [{"prmm": "mm"}])
def test_read_emep_clean_filepaths_error(data_path: Path, years, freq: str, error: str):
    reader = ReadMscwCtm(data_dir=str(data_path))
    tst = reader.FREQ_CODES[freq]
    filepaths = reader._filepaths
    with pytest.raises(ValueError) as e:
        reader._clean_filepaths(filepaths, years, tst)

    assert error in str(e.value)


@pytest.mark.parametrize(
    "freq,wrong_name",
    [
        ("day", "Base_day.nc"),
        ("month", "Base_month.nc"),
    ],
)
@pytest.mark.parametrize("vars_and_units", [{"prmm": "mm"}])
def test_read_emep_wrong_filenames(data_path: Path, freq: str, wrong_name: str):
    reader = ReadMscwCtm(data_dir=str(data_path))
    tst = reader.FREQ_CODES[freq]
    filepaths = reader._filepaths
    years = reader._get_yrs_from_filepaths()
    cleaned_paths = reader._clean_filepaths(filepaths, years, tst)

    wrong_path = Path(filepaths[0]).with_name(wrong_name)
    filepaths[0] = str(wrong_path)
    new_years = reader._get_yrs_from_filepaths()
    new_cleaned_paths = reader._clean_filepaths(filepaths, new_years, tst)

    assert len(years) == len(new_years)
    assert len(cleaned_paths) == len(new_cleaned_paths)


@pytest.mark.parametrize(
    "freq,wrong_tst",
    [
        ("month", "minute"),
        ("day", "daily"),
    ],
)
@pytest.mark.parametrize("vars_and_units", [{"prmm": "mm"}])
def test_read_emep_wrong_tst(data_path: Path, wrong_tst: str):
    reader = ReadMscwCtm(data_dir=str(data_path))
    with pytest.raises(ValueError) as e:
        filepaths = reader._filepaths
        wrong_path = Path(filepaths[0]).with_name(f"Base_{wrong_tst}.nc")
        reader._get_tst_from_file(str(wrong_path))

    assert " does not match file_pattern " in str(e.value)


def test_read_emep_LF_tst(tmp_path: Path):
    data_path = emep_data_path(tmp_path, "month", vars_and_units={"prmm": "mm"})
    reader = ReadMscwCtm(data_dir=str(data_path))
    with pytest.raises(ValueError) as e:
        filepaths = reader._filepaths
        wrong_path = Path(filepaths[0]).with_name("Base_LF_month.nc")
        reader._get_tst_from_file(str(wrong_path))

    assert " does not match file_pattern " in str(e.value)


@pytest.mark.xfail
def test_read_emep_year_defined_twice(tmp_path: Path):
    data_path = emep_data_path(tmp_path, "day", vars_and_units={"prmm": "mm"})
    reader = ReadMscwCtm(data_dir=str(data_path))
    filepaths = reader._filepaths
    wrong_path = Path(filepaths[0]).with_name("Base_day.nc")
    filepaths.append(str(wrong_path))
    new_yrs = reader._get_yrs_from_filepaths()
    with pytest.raises(ValueError) as e:
        reader._clean_filepaths(filepaths, new_yrs, "daily")
    assert "is already found:" in str(e.value)


@pytest.mark.parametrize(
    "year,freq,num",
    [
        ("", "day", 5),
        ("2017", "month", 1),
        ("2019", "hour", 1),
    ],
)
@pytest.mark.parametrize("vars_and_units", [{"prmm": "mm"}])
def test_read_emep_multiple_dirs(data_path: Path, year: str, freq: str, num: int):
    reader = ReadMscwCtm(data_dir=str(data_path / year))
    assert len(reader._filepaths) == num
    assert all(freq in Path(file).stem for file in reader._filepaths)

    tst = reader.FREQ_CODES[freq]
    data = reader.read_var("prmm", ts_type=tst)
    assert data.ts_type == tst


def test_read_emep_multiple_dirs_hour_error(tmp_path: Path):
    data_path = emep_data_path(tmp_path, "hour", vars_and_units={"prmm": "mm"})
    reader = ReadMscwCtm(data_dir=str(data_path))
    with pytest.raises(ValueError) as e:
        reader.read_var("prmm", ts_type="hourly")
    assert "can not be hourly when using multiple years" in str(e.value)


@pytest.mark.parametrize(
    "year,freq,num",
    [
        ("", "day", 5),
        ("2017", "month", 1),
        ("2019", "hour", 1),
    ],
)
@pytest.mark.parametrize("vars_and_units", [{"prmm": "mm"}])
def test_search_all_files(data_path: Path, year: str, num: int):
    reader = ReadMscwCtm()
    with pytest.raises(AttributeError):
        reader._search_all_files()

    with pytest.raises(AttributeError):
        reader._filepaths

    reader._data_dir = str(data_path / year)

    reader._search_all_files()
    assert len(reader._filepaths) == num


@pytest.mark.parametrize(
    "year,freq,",
    [
        ("", ["day"]),
        ("", ["day", "hour"]),
        ("2017", ["month"]),
        ("2019", ["hour", "day", "month"]),
    ],
)
@pytest.mark.parametrize("vars_and_units", [{"prmm": "mm"}])
def test_ts_types(data_path: Path, year: str, freq: list[str]):
    reader = ReadMscwCtm()
    with pytest.raises(AttributeError):
        reader.ts_types

    reader._data_dir = str(data_path / year)
    ts_types = reader.ts_types
    assert len(ts_types) == len(freq)


def test_add_aux_compute(tmp_path: Path):
    data_path = emep_data_path(
        tmp_path, "day", vars_and_units={"concno3c": "ug m-3", "concno3f": "ug m-3"}
    )
    reader = ReadMscwCtm(data_dir=str(data_path / "2017"))

    def calc_concno3(concno3c, concno3f):
        concno3 = concno3c.copy(deep=True) + concno3f.copy(deep=True)
        concno3.attrs["units"] = "ug m-3"
        return concno3

    new_var_name = "concno3"
    vars_required = ["concno3c", "concno3f"]
    func = calc_concno3

    reader.add_aux_compute(new_var_name, vars_required=vars_required, fun=func)

    assert reader.has_var(new_var_name)

    data = reader.read_var(new_var_name, "daily")

    assert data.var_name == new_var_name


def test_emep_vars():
    new_var_name = "concno3"
    new_mapping = "SURF_ug_NO3_C"

    reader = ReadMscwCtm(emep_vars={new_var_name: new_mapping})

    assert new_var_name in reader.var_map
    assert reader.var_map[new_var_name] == new_mapping


def test_reader_regexp(tmp_path: Path):
    path = tmp_path / "2022"
    path.mkdir()
    # Create files that should be matched by regexp.
    for i in range(10):
        (path / f"file_{i}.nc").touch()

    # Create files that should NOT be matched by regexp.
    for i in range(10):
        (path / f"blah_{i}.nc").touch()

    reader = ReadMscwCtm("test", data_dir=str(path), file_pattern=r"^file_\d\.nc$")

    reader._search_all_files()

    assert len(reader._private.filepaths) == 10


def test_reader_read_deposition():
    # https://github.com/metno/pyaerocom/issues/1624
    file_path = Path("~/MyPyaerocom/testdata-minimal/modeldata/EMEP_DEPOSITION/").expanduser()
    reader = ReadMscwCtm(data_dir=str(file_path))
    reader.read_var("dryrdn", ts_type="monthly")



import glob
import logging

import aerovaldb
import xarray as xr

from pyaerocom import ColocatedData, GriddedData, TsType, __version__, const
from pyaerocom.aeroval._processing_base import DataImporter, ProcessingEngine
from pyaerocom.aeroval.json_utils import round_floats
from pyaerocom.aeroval.modelmaps_helpers import (
    CONTOUR,
    OVERLAY,
    _jsdate_list,
    calc_contour_json,
    find_netcdf_files,
    plot_overlay_pixel_maps,
)
from pyaerocom.aeroval.varinfo_web import VarinfoWeb
from pyaerocom.colocation.colocator import Colocator
from pyaerocom.exceptions import (
    DataCoverageError,
    DataDimensionError,
    DataQueryError,
    EntryNotAvailable,
    ModelVarNotAvailable,
    TemporalResolutionError,
    VariableDefinitionError,
    VarNotAvailableError,
)
from pyaerocom.units.helpers import get_standard_unit


logger = logging.getLogger(__name__)

MODELREADERS_USE_MAP_FREQ = ["ReadMscwCtm"]  # , "ReadCAMS2_83"]


class ModelMapsEngine(ProcessingEngine, DataImporter):
    """
    Engine for processing of model maps
    """

    def _get_run_kwargs(self, **kwargs):
        try:
            model_list = kwargs["model_list"]
        except KeyError:
            model_list = self.cfg.model_cfg.keylist()
        try:
            var_list = kwargs["var_list"]
        except Exception:
            var_list = None
        return model_list, var_list

    def run(self, **kwargs):
        model_list, var_list = self._get_run_kwargs(**kwargs)

        for model in model_list:
            try:
                self._run_model(model, var_list)
            except VarNotAvailableError:
                logger.warning(f"no data for model {model}, skipping")
                continue

        self.cfg.modelmaps_opts.maps_freq = (
            self._get_maps_freq()
        )  # if needed, reassign "coarsest" to actual coarsest frequency

    def _get_vars_to_process(self, model_name, var_list):
        mvars = self.cfg.model_cfg.get_entry(model_name).get_vars_to_process(
            self.cfg.obs_cfg.get_all_vars()
        )[1]
        all_vars = sorted(list(set(mvars)))
        if var_list is not None:
            all_vars = [var for var in var_list if var in all_vars]
        return all_vars

    def _get_obs_vars_to_process(self, obs_name, var_list):
        ovars = self.cfg.obs_cfg.get_entry(obs_name).get_all_vars()
        all_vars = sorted(list(set(ovars)))
        if var_list is not None:
            all_vars = [var for var in var_list if var in all_vars]
        return all_vars

    def _run_model(self, model_name: str, var_list):
        """Run evaluation of map processing

        Create json files for model-maps display. This analysis does not
        require any observation data but processes model output at all model
        grid points, which is then displayed on the website in the maps
        section.

        Parameters
        ----------
        model_name : str
            name of model to be processed
        var_list : list, optional
            name of variable to be processed. If None, all available
            observation variables are used.

        """

        try:
            var_list = self._get_vars_to_process(model_name, var_list)
        except EntryNotAvailable:
            var_list = self._get_obs_vars_to_process(model_name, var_list)
        if not var_list:
            raise VarNotAvailableError("List of variables is empty.")
        for var in var_list:
            logger.info(f"Processing model maps for {model_name} ({var})")

            try:  # pragma: no cover
                make_contour, make_overlay = False, False
                if isinstance(self.cfg.modelmaps_opts.plot_types, dict):
                    make_contour = CONTOUR in self.cfg.modelmaps_opts.plot_types.get(
                        model_name, False
                    )
                    make_overlay = OVERLAY in self.cfg.modelmaps_opts.plot_types.get(
                        model_name, False
                    )
                if CONTOUR in self.cfg.modelmaps_opts.plot_types or make_contour:
                    self._process_contour_map_var(model_name, var, self.reanalyse_existing)

                if OVERLAY in self.cfg.modelmaps_opts.plot_types or make_overlay:
                    # create overlay (pixel) plots
                    self._process_overlay_map_var(model_name, var, self.reanalyse_existing)

            except ModelVarNotAvailable as ex:
                logger.warning(f"{ex}")
            except (
                TemporalResolutionError,
                DataCoverageError,
                VariableDefinitionError,
                DataQueryError,
            ) as e:
                if self.raise_exceptions:
                    raise
                logger.warning(f"Failed to process maps for {model_name} {var} data. Reason: {e}.")
        return

    def _check_dimensions(self, data: GriddedData) -> "GriddedData":
        if not data.has_latlon_dims:
            raise DataDimensionError("data needs to have latitude an longitude dimension")
        elif not data.has_time_dim:
            raise DataDimensionError("data needs to have latitude an longitude dimension")
        if data.ndim == 4:
            data = data.extract_surface_level()
        return data

    def _process_contour_map_var(self, model_name, var, reanalyse_existing):  # pragma: no cover
        """
        Process model data to create map geojson files

        Parameters
        ----------
        model_name : str
            name of model
        var : str
            name of variable
        reanalyse_existing : bool
            if True, already existing json files will be reprocessed

        Raises
        ------
        ValueError
            If vertical code of data is invalid or not set
        AttributeError
            If the data has the incorrect number of dimensions or misses either
            of time, latitude or longitude dimension.
        ModelVarNotAvailable
            If model/var data cannot be read
        """
        try:
            data = self._read_model_data(model_name, var)
        except Exception as e:
            raise ModelVarNotAvailable(
                f"Cannot read data for model {model_name} (variable {var}): {e}"
            )

        var_ranges_defaults = self.cfg.var_scale_colmap
        if var in var_ranges_defaults.keys():
            cmapinfo = var_ranges_defaults[var]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])
        else:
            cmapinfo = var_ranges_defaults["default"]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])

        data = self._check_dimensions(data)

        freq = self._get_maps_freq()
        tst = TsType(data.ts_type)

        if tst < freq:
            raise TemporalResolutionError(f"need {freq} or higher, got{tst}")
        elif tst > freq:
            data = data.resample_time(str(freq))

        data.check_unit()

        if not reanalyse_existing:
            # check if all files have already been produced
            # if even just one is missing, all is gonna be recomputed
            ts = _jsdate_list(data)

            uris_contour = self.avdb.query(
                aerovaldb.routes.Route.CONTOUR_TIMESPLIT,
                project=self.exp_output.proj_id,
                experiment=self.exp_output.exp_id,
            )
            all_times = [int(uri.meta["timestep"]) for uri in uris_contour]

            if all([date in all_times for date in ts]):
                logger.info(
                    f"Skipping contour processing of {var}_{model_name}: data already exists {uris_contour}."
                )
                return

        # first calculate and save geojson with contour levels
        contourjson = calc_contour_json(data, cmap=varinfo.cmap, cmap_bins=varinfo.cmap_bins)

        with self.avdb.lock():
            for time, contour in contourjson.items():
                self.avdb.put_contour(
                    contour,
                    self.exp_output.proj_id,
                    self.exp_output.exp_id,
                    self.cfg.model_cfg.get_entry(model_name).model_rename_vars.get(var, var),
                    model_name,
                    timestep=time,
                )

    def _process_overlay_map_var(self, model_name, var, reanalyse_existing):  # pragma: no cover
        """Process overlay map (pixels) for either model or obserations
        argument model_name is a misnomer because this can also be applied to observation networks

        Args:
            model_name (str): name of model or obs to make overlay pixel maps of
            var (str): variable name
        """

        if self.cfg.processing_opts.only_json:  # we have colocated data
            data = self._process_only_json(model_name, var)
        else:
            try:
                data = self.read_gridded_obsdata(model_name, var)
            except EntryNotAvailable:
                try:
                    data = self._read_model_data(model_name, var)
                except Exception as e:
                    raise ModelVarNotAvailable(
                        f"Cannot read data for model {model_name} (variable {var}): {e}"
                    )

        var_ranges_defaults = self.cfg.var_scale_colmap

        if var in var_ranges_defaults.keys():
            cmapinfo = var_ranges_defaults[var]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])
        else:
            cmapinfo = var_ranges_defaults["default"]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])

        if not self.cfg.processing_opts.only_json:
            data = self._check_dimensions(data)

        freq = self._get_maps_freq()

        tst = TsType(data.ts_type)

        if tst < freq:
            raise TemporalResolutionError(f"need {freq} or higher, got{tst}")
        elif tst > freq:
            if isinstance(data, GriddedData):
                data = data.resample_time(str(freq))
            elif isinstance(data, xr.DataArray):
                data = data.resample(time=str(freq)[0].capitalize()).mean()

        ts = _jsdate_list(data)
        if isinstance(data, GriddedData):
            data.check_unit()
            data = data.to_xarray().load()

        if self.cfg.processing_opts.only_model_maps:
            self._check_ts_for_only_model_maps(model_name, var, ts, data)

        for i, date in enumerate(ts):
            try:
                write_var_name = self.cfg.model_cfg.get_entry(model_name).model_rename_vars.get(
                    var, var
                )
            except EntryNotAvailable:
                write_var_name = var

            # Note this should match the output location defined in aerovaldb
            overlay_uris = self.avdb.query(
                aerovaldb.routes.Route.MAP_OVERLAY,
                project=self.exp_output.proj_id,
                experiment=self.exp_output.exp_id,
            )

            if not reanalyse_existing:
                if any(
                    uri.meta["variable"] == write_var_name
                    and uri.meta["source"] == model_name
                    and uri.meta["date"] == date
                    for uri in overlay_uris
                ):
                    logger.info(
                        f"Skipping overlay processing for model={model_name}, var={write_var_name}, date={date}: data already exists."
                    )
                    continue

            overlay_plot = plot_overlay_pixel_maps(
                data[i],
                cmap="gray",
                cmap_bins=varinfo.cmap_bins,
                format=self.cfg.modelmaps_opts.overlay_save_format,
            )

            with self.avdb.lock():
                self.avdb.put_map_overlay(
                    overlay_plot,
                    self.exp_output.proj_id,
                    self.exp_output.exp_id,
                    model_name,
                    write_var_name,
                    str(date),
                )

    def _get_maps_freq(self) -> TsType:
        """
        Gets the maps reading frequency. If maps_freq in cfg is coarsest, it takes the coarsest
        of the given frequencies. Else it just returns the maps_freq

        Returns
        -------
        TsType
        """
        maps_freq = TsType(self.cfg.modelmaps_opts.maps_freq)
        if maps_freq == "coarsest":  # TODO: Implement this in terms of a TsType object. #1267
            freq = min(TsType(fq) for fq in self.cfg.time_cfg.freqs)
            freq = min(freq, self.cfg.time_cfg.main_freq)
        else:
            freq = maps_freq
        return freq

    def _get_read_model_freq(self, model_ts_types: list) -> TsType:
        """
        Tries to find the best TsType to read. Checks for available ts types with the following priority

        1. If the freq from _get_maps_freq is available
        2. If maps_freq is explicitly given, and is available
        3. Iterates through the freqs given in the config, and find the coarsest available ts type

        Raises
        -------

        ValueError
            If no ts types are possible to read

        Returns
        -------
        TSType
        """
        wanted_freq = self._get_maps_freq()
        if wanted_freq in model_ts_types:
            return wanted_freq

        maps_freq = TsType(self.cfg.modelmaps_opts.maps_freq)

        if maps_freq != "coarsest":
            if maps_freq not in model_ts_types:
                raise ValueError(
                    f"Could not find any model data for given maps_freq. {maps_freq} is not in {model_ts_types}"
                )
            return maps_freq

        for freq in sorted(TsType(fq) for fq in self.cfg.time_cfg.freqs):
            if freq in model_ts_types:
                logger.info(f"Found coarsest maps_freq that is available as model data: {freq}")
                return freq

        freq = min(TsType(fq) for fq in model_ts_types)
        logger.info(f"Found coarsest freq available as model data: {freq}")
        return freq

    def _read_model_data(self, model_name: str, var: str) -> GriddedData:
        """
        Function for reading the model data without going through the colocation object.
        This means that none of the checks normally done in the colocation class are run.

        Parameters
        ----------
        model_name : str
            name of model
        var : str
            name of variable

        Returns
        -----------
        Griddeddata
            the read data
        """
        start, stop = self.cfg.colocation_opts.start, self.cfg.colocation_opts.stop
        if self.cfg.colocation_opts.model_use_climatology:
            # overwrite start and stop to read climatology file for model
            start, stop = 9999, None

        data_id = self.cfg.model_cfg.get_entry(model_name).model_id

        try:
            data_dir = self.cfg.model_cfg.get_entry(model_name).model_data_dir
        except Exception as e:
            logger.info(f"Could not find model dir. Setting to None. Error {str(e)}")
            data_dir = None

        try:
            model_reader = self.cfg.model_cfg.get_entry(model_name).gridded_reader_id["model"]
        except Exception as e:
            logger.info(f"Could not find model reader. Setting to None. Error {str(e)}")
            model_reader = None

        if model_reader is not None:
            reader_class = Colocator.SUPPORTED_GRIDDED_READERS[model_reader]
        else:
            reader_class = Colocator.SUPPORTED_GRIDDED_READERS["ReadGridded"]

        reader = reader_class(
            data_id=data_id,
            data_dir=data_dir,
            **self.cfg.colocation_opts.model_kwargs,
        )

        if var in self.cfg.model_cfg.get_entry(model_name).model_read_aux:
            aux_instructions = self.cfg.model_cfg.get_entry(model_name).model_read_aux[var]
            reader.add_aux_compute(var_name=var, **aux_instructions)

        kwargs = {}
        kwargs.update(**self.cfg.colocation_opts.model_kwargs)
        if var in self.cfg.colocation_opts.model_read_opts:
            kwargs.update(self.cfg.colocation_opts.model_read_opts[var])
        kwargs.update(self.cfg.get_model_entry(model_name).get("model_kwargs", {}))

        if model_reader is not None and model_reader in MODELREADERS_USE_MAP_FREQ:
            ts_types = reader.ts_types
            ts_type_read = str(self._get_read_model_freq(ts_types))
        else:
            model_ts_type_read = self.cfg.model_cfg.get_entry(model_name).model_ts_type_read
            if model_ts_type_read:
                ts_type_read = model_ts_type_read
            else:
                ts_type_read = (
                    self.cfg.colocation_opts.ts_type
                )  # emulates the old way closer than None

        data = reader.read_var(
            var,
            start=start,
            stop=stop,
            ts_type=ts_type_read,
            vert_which=self.cfg.colocation_opts.obs_vert_type,
            flex_ts_type=self.cfg.colocation_opts.flex_ts_type,
            **kwargs,
        )

        rm_outliers = self.cfg.colocation_opts.model_remove_outliers
        outlier_ranges = self.cfg.colocation_opts.model_outlier_ranges

        if rm_outliers:
            if var in outlier_ranges:
                low, high = outlier_ranges[var]
            else:
                var_info = const.VARS[var]
                low, high = var_info.minimum, var_info.maximum

            data.remove_outliers(low, high, inplace=True)

        data.convert_unit(get_standard_unit(data.var_name))
        return data

    def _check_ts_for_only_model_maps(
        self, name: str, var: str, dates: list[int], data: xr.Dataset
    ):  # pragma: no cover
        maps_freq = str(self._get_maps_freq())
        if name in self.cfg.obs_cfg.keylist():
            with self.avdb.lock():
                timeseries = self.avdb.get_timeseries(
                    self.cfg.proj_info.proj_id,
                    self.cfg.exp_info.exp_id,
                    "ALL",
                    name,
                    var,
                    self.cfg.obs_cfg.get_entry(name).obs_vert_type,
                    default={},
                )

                for model_name in self.cfg.model_cfg.keylist():
                    if (
                        model_name in timeseries
                        and timeseries[model_name].get(maps_freq + "_mod", False)
                        and timeseries[model_name].get(maps_freq + "_obs", False)
                    ):
                        continue

                    timeseries.setdefault(model_name, {})
                    timeseries[model_name].setdefault(maps_freq + "_date", dates)
                    timeseries[model_name].setdefault(
                        maps_freq + "_obs",
                        list(round_floats(data.mean(dim=("latitude", "longitude")).values)),
                    )

                    timeseries[model_name].setdefault("obs_var", var)
                    if hasattr(data, "units"):
                        timeseries[model_name].setdefault("obs_unit", data.units)
                    elif hasattr(data, "var_units"):
                        timeseries[model_name].setdefault("obs_unit", data.var_units[1])
                    else:
                        raise ValueError("Can not determine obs units")
                    timeseries[model_name].setdefault("obs_name", name)
                    timeseries[model_name].setdefault(
                        "var_name_web", self.cfg.obs_cfg.get_web_interface_name(name)
                    )
                    timeseries[model_name].setdefault(
                        "vert_code", self.cfg.obs_cfg.get_entry(name).obs_vert_type
                    )
                    timeseries[model_name].setdefault("obs_freq_src", maps_freq)

                self.avdb.put_timeseries(
                    timeseries,
                    self.cfg.proj_info.proj_id,
                    self.cfg.exp_info.exp_id,
                    "ALL",
                    name,
                    var,
                    self.cfg.obs_cfg.get_entry(name).obs_vert_type,
                )
        elif name in self.cfg.model_cfg.keylist():
            with self.avdb.lock():
                for obs_name in self.cfg.obs_cfg.keylist():
                    timeseries = self.avdb.get_timeseries(
                        self.cfg.proj_info.proj_id,
                        self.cfg.exp_info.exp_id,
                        "ALL",
                        obs_name,
                        var,
                        self.cfg.obs_cfg.get_entry(obs_name).obs_vert_type,
                        default={},
                    )

                    if (
                        name in timeseries
                        and timeseries[name].get(maps_freq + "_mod", False)
                        and timeseries[name].get(maps_freq + "_obs", False)
                    ):
                        continue

                    timeseries.setdefault(name, {})
                    timeseries[name].setdefault(maps_freq + "_date", dates)
                    timeseries[name].setdefault(
                        maps_freq + "_mod",
                        list(round_floats(data.mean(dim=("latitude", "longitude")).values)),
                    )
                    timeseries[name].setdefault("station_name", "ALL")
                    timeseries[name].setdefault("pyaerocom_version", __version__)
                    timeseries[name].setdefault("mod_var", var)
                    if hasattr(data, "units"):
                        timeseries[name].setdefault("mod_unit", data.units)
                    elif hasattr(data, "var_units"):
                        timeseries[name].setdefault("mod_unit", data.var_units[0])
                    else:
                        raise ValueError("Can not determine model units")
                    timeseries[name].setdefault("model_name", name)
                    timeseries[name].setdefault("mod_freq_src", maps_freq)

                    self.avdb.put_timeseries(
                        timeseries,
                        self.cfg.proj_info.proj_id,
                        self.cfg.exp_info.exp_id,
                        "ALL",
                        obs_name,
                        var,
                        self.cfg.obs_cfg.get_entry(obs_name).obs_vert_type,
                    )
        else:
            raise ValueError(
                f"{name=} not is not in either {self.cfg.obs_cfg.keylist()=} nor {self.cfg.model_cfg.keylist()=}"
            )

    def _process_only_json(self, model_name, var):  # pragma: no cover
        """Process data from ColocatedData for overlay map for if only_json = True."""
        try:
            preprocessed_coldata_dir = glob.escape(
                self.cfg.model_cfg.get_entry(model_name).model_data_dir
            )
            mask = f"{preprocessed_coldata_dir}/*.nc"
            file_to_convert = glob.glob(mask)
        except KeyError:
            preprocessed_coldata_dir = glob.escape(
                self.cfg.obs_cfg.get_entry(model_name).coldata_dir
            )
            mask = f"{preprocessed_coldata_dir}/{model_name}/*.nc"
            matching_files = find_netcdf_files(
                directory=preprocessed_coldata_dir, strings=[model_name, var]
            )

            if len(matching_files) > 1:
                logger.info(
                    f"Found more than one colocated data file for {model_name=} {var=}. Using the first one found - this theoretically should be consistent across files."
                )
            file_to_convert = matching_files[:1]

        if len(file_to_convert) != 1:
            raise ValueError(
                "Can only handle one colocated data object for plotting for a given (model, obs, var). "
                "Note that when providing a colocated data object, it must be provided via the model_data_dir argument in a ModelEntry instance. "
                "It must also be provided via the coldata_dir argument in the ObsEntry instance. "
                "Additionally, note that the coldatadir does not contain the model_name at the end of the directory, "
                "whereas the coldata_dir does not."
            )

        coldata = ColocatedData(data=file_to_convert[0])
        data = coldata.data.sel(data_source=model_name)
        data = data.drop_vars("data_source")
        data = data.transpose("time", "latitude", "longitude")
        data = data.sortby(["latitude", "longitude"])
        return data



import codecs
import logging
import os
from glob import glob

import numpy as np
import pandas as pd
from tqdm import tqdm

from pyaerocom.exceptions import DataRetrievalError
from pyaerocom.io import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadAirNow(ReadUngriddedBase):
    """
    Reading routine for North-American Air Now observations
    """

    # Data type of files
    _FILETYPE = ".dam"

    # File search mask to recursively retrieve list of data files
    _FILEMASK = f"/monthly/*{_FILETYPE}"

    #: Version log of this class (for caching)
    __version__ = "0.12"

    #: Column delimiter
    FILE_COL_DELIM = "|"

    #: Columns in data files
    FILE_COL_NAMES = [
        "date",
        "time",
        "station_id",
        "station_name",
        "time_zone",
        "variable",
        "unit",
        "value",
        "institute",
    ]
    # will be used millions of times, therefore store it
    FILE_COL_ROW_NUMBER = len(FILE_COL_NAMES)

    # row # for the variable
    ROW_VAR_COL = 5

    #: Mapping of columns in station metadata file to pyaerocom standard
    STATION_META_MAP = {
        "aqsid": "station_id",
        "name": "station_name",
        "lat": "latitude",
        "lon": "longitude",
        "elevation": "altitude",
        "city": "city",
        "address": "address",
        "timezone": "timezone",
        "environment": "area_classification",
        "populationclass": "station_classification",
        "modificationdate": "modificationdate",
        "comment": "comment",
    }

    #: conversion functions for metadata dtypes
    STATION_META_DTYPES = {
        "station_id": str,
        "station_name": str,
        "latitude": float,
        "longitude": float,
        "altitude": float,
        "city": str,
        "address": str,
        "timezone": str,
        "area_classification": str,
        "station_classification": str,
        "modificationdate": str,
        "comment": str,
    }

    # strings to be replaced in original station names
    REPLACE_STATNAME = {"&": "and", "/": " ", ":": " ", ".": " ", "'": ""}

    # Years in timestamps in the files are 2-digit (e.g. 20 for 2020)
    BASEYEAR = 2000

    #: Name of dataset (OBS_ID)
    DATA_ID = "AirNow"

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: Units found in data files
    UNIT_MAP = {
        "C": "celsius",
        "M/S": "m s-1",
        "MILLIBAR": "mbar",
        "MM": "mm",
        "PERCENT": "%",
        "PPB": "ppb",
        "PPM": "ppm",
        "UG/M3": "ug m-3",
        "WATTS/M2": "W m-2",
    }

    #: Variable names in data files
    VAR_MAP = {
        "concbc": "BC",
        "concpm10": "PM10",
        "concpm25": "PM2.5",
        "vmrco": "CO",
        "vmrnh3": "NH3",
        "vmrno": "NO",
        "vmrno2": "NO2",
        "vmrnox": "NOX",
        "vmrnoy": "NOY",
        "vmro3": "OZONE",
        "vmrso2": "SO2",
    }

    #: List of variables that are provided
    PROVIDES_VARIABLES = list(VAR_MAP)

    #: Default variables
    DEFAULT_VARS = PROVIDES_VARIABLES

    #: Frequency of measurements
    TS_TYPE = "hourly"

    #: file containing station metadata
    STAT_METADATA_FILENAME = "allStations_20191224.csv"

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id=data_id, data_dir=data_dir)
        self.make_datetime64_array = np.vectorize(self._date_time_str_to_datetime64)
        self._station_metadata = None

    @property
    def station_metadata(self):
        """Dictionary containing global metadata for each site"""
        if self._station_metadata is None:
            self._init_station_metadata()
        return self._station_metadata

    def _date_time_str_to_datetime64(self, date, time):
        """
        Convert date and time string into datetime64 object

        Parameters
        ----------
        date : str
            date string as mm/dd/yy as in data files
        time : str
            time of the day as HH:MM

        Returns
        -------
        datetime64[s]
        """
        mm, dd, yy = date.split("/")
        yr = str(self.BASEYEAR + int(yy))
        # returns as datetime64[s]
        return np.datetime64(f"{yr}-{mm}-{dd}T{time}:00")

    def _read_metadata_file(self):
        """
        Read station metadatafile

        Returns
        -------
        cfg : pandas.DataFrame
            metadata dataframe

        """
        fn = os.path.join(self.data_dir, self.STAT_METADATA_FILENAME)
        cfg = pd.read_csv(fn, sep=",", converters={"aqsid": lambda x: str(x)})
        return cfg

    def _correct_station_name(self, station_name):
        """
        Remove unwanted chars from original station names

        Parameters
        ----------
        station_name : str
            original station name

        Returns
        -------
        str
            station name cleaned of chars defined in :attr:`REPLACE_STATNAME`

        """
        for search, replace in self.REPLACE_STATNAME.items():
            station_name = station_name.replace(search, replace)
        return station_name

    def _init_station_metadata(self):
        """
        Initiate metadata for all stations

        Returns
        -------
        dict
            dictionary with metadata dictionaries for all stations

        """
        cfg = self._read_metadata_file()
        meta_map = self.STATION_META_MAP

        cols = list(cfg.columns.values)
        col_idx = {}
        for from_meta, to_meta in meta_map.items():
            col_idx[to_meta] = cols.index(from_meta)

        arr = cfg.values
        dtypes = self.STATION_META_DTYPES
        stats = {}
        for row in arr:
            stat = {}
            for meta_key, col_num in col_idx.items():
                stat[meta_key] = dtypes[meta_key](row[col_num])
            sid = stat["station_id"]

            stat["station_name"] = self._correct_station_name(stat["station_name"])
            stat["data_id"] = self.data_id
            stat["ts_type"] = self.TS_TYPE
            stats[sid] = stat
        self._station_metadata = stats
        return stats

    def get_file_list(self):
        """
        Retrieve list of data files

        Returns
        -------
        list
        """
        basepath = self.data_dir
        pattern = f"{basepath}{self._FILEMASK}"
        files = sorted(glob(pattern))
        return files

    def _read_file(self, file):
        """
        Read one datafile using :func:`pandas.read_csv`

        Parameters
        ----------
        file : str
            file path

        Returns
        -------
        df : pandas.DataFrame
            DataFrame containing the file data

        """

        # try utf_8 and cp863 reading first, then
        # determine file encoding and provide that to pandas
        # just determining the encoding is too slow given the # of files
        # Airbase consists of
        # just trying a couple of encodings and not determining the encoding all
        # the time speeds up reading by a factor of 5
        # make sure certain dtypes are set for a few rows (e.g. station code as str)
        try:
            encoding = "utf_8"
            df = pd.read_csv(
                file,
                sep=self.FILE_COL_DELIM,
                names=self.FILE_COL_NAMES,
                encoding=encoding,
                on_bad_lines="skip",
                dtype={2: str, 4: float, 7: float},
            )
        except UnicodeDecodeError:
            encoding = "cp863"
            df = pd.read_csv(
                file,
                sep=self.FILE_COL_DELIM,
                names=self.FILE_COL_NAMES,
                encoding=encoding,
                on_bad_lines="skip",
                dtype={2: str, 4: float, 7: float},
            )
        except Exception:
            encoding = self.get_file_encoding(file)
            df = pd.read_csv(
                file,
                sep=self.FILE_COL_DELIM,
                names=self.FILE_COL_NAMES,
                encoding=encoding,
                on_bad_lines="skip",
                dtype={
                    2: str,
                    4: float,
                    7: float,
                },
            )
        return df

    def _read_files(self, files, vars_to_retrieve):
        """
        Read input variables from list of files

        Parameters
        ----------
        files : list
            list of data files
        vars_to_retrieve : list
            list of variables to retrieve

        Raises
        ------
        NotImplementedError
            if several timezones are assigned to the same station
        AttributeError
            if data unit is unknown

        Returns
        -------
        stats : list
            list of StationData objects

        """
        logger.info("Read AirNow data file(s)")
        file_vars_to_retrieve = [self.VAR_MAP[x] for x in vars_to_retrieve]

        arrs = []
        unique_stat_ids = None
        for i in tqdm(range(len(files)), disable=None):
            fp = files[i]
            filedata = self._read_file(fp)
            for i, filevar in enumerate(file_vars_to_retrieve):
                arrs.append(filedata[filedata["variable"] == filevar].values)
                if unique_stat_ids is None:
                    unique_stat_ids = np.unique(
                        arrs[-1][:, self.FILE_COL_NAMES.index("station_id")]
                    )
                else:
                    try:
                        unique_stat_ids = np.union1d(
                            unique_stat_ids,
                            np.unique(arrs[-1][:, self.FILE_COL_NAMES.index("station_id")]),
                        )
                    except (ValueError, TypeError):
                        print(arrs[-1][:, self.FILE_COL_NAMES.index("station_id")])
                        raise DataRetrievalError(
                            f"file {fp}: error in creating unique stationlist"
                        )
        if len(arrs) == 0:
            raise DataRetrievalError("None of the input variables could be found in input list")
        return self._filedata_to_statlist(arrs, vars_to_retrieve, unique_stat_ids=unique_stat_ids)

    def _filedata_to_statlist(
        self,
        arrs: list[np.ndarray],
        vars_to_retrieve: list[str],
        unique_stat_ids: list[str] = None,
    ) -> list[StationData]:
        """
        Convert loaded filedata into list of StationData objects

        Parameters
        ----------
        arrs : list
            list of numpy arrays extracted from each file
            (see :func:`_read_files`).
        vars_to_retrieve : list
            list of variables to be retrieved from input data.


        Returns
        -------
        stats : list
            list of :class:`StationData` objects, one for each var and station.
            :type unique_stat_ids: object

        """
        # doubling of RAM usage!
        data = np.concatenate(arrs)
        # so kill the input data right afterwards
        arrs = None

        logger.info("Converting filedata to list of StationData")
        stat_meta = self.station_metadata
        stat_ids = list(stat_meta)
        varcol = self.FILE_COL_NAMES.index("variable")
        statcol = self.FILE_COL_NAMES.index("station_id")
        tzonecol = self.FILE_COL_NAMES.index("time_zone")
        unitcol = self.FILE_COL_NAMES.index("unit")
        valcol = self.FILE_COL_NAMES.index("value")

        stats = []
        # shortcut for limited testing dataset because the code
        # below the try statement fails with testing:
        # E               ValueError: cannot call `vectorize` on size 0 inputs unless `otypes` is set
        # ../../../mambaforge/envs/pyadev-applied/lib/python3.11/site-packages/numpy/lib/function_base.py:2363: ValueError
        try:
            dtime = self.make_datetime64_array(data[:, 0], data[:, 1])
        except ValueError:
            raise DataRetrievalError("None of the input variables could be found in input list")

        for var in vars_to_retrieve:
            # extract only variable data (should speed things up)
            var_in_file = self.VAR_MAP[var]
            mask = data[:, varcol] == var_in_file
            # another RAM consuming op, the mask can just be used all the time (for the price of readability...)
            subset = data[mask]
            dtime_subset = dtime[mask]
            # there are stations with a all numeric station ID, but type hints in pd.read_csv made sure
            # they are read as str...
            if unique_stat_ids is None:
                statlist = np.unique(subset[:, statcol])
            else:
                statlist = unique_stat_ids
            for stat_id in tqdm(statlist, desc=var, disable=None):
                if stat_id not in stat_ids:
                    continue
                statmask = subset[:, statcol] == stat_id
                if statmask.sum() == 0:
                    continue
                # RAM consuming op...
                statdata = subset[statmask]
                timestamps = dtime_subset[statmask]
                # timezone offsets (there's a half hour time zone!, so float)
                # not sure why the astype(float) is needed in between, but it does not work without...
                toffs = statdata[:, tzonecol].astype(float).astype("timedelta64[h]")
                timestamps += toffs
                stat = StationData(**stat_meta[stat_id])

                vals = statdata[:, valcol]
                units = np.unique(statdata[:, unitcol])
                # errors that did not occur in v0 but that may occur
                assert len(units) == 1
                assert units[0] in self.UNIT_MAP
                stat["dtime"] = timestamps
                stat["timezone"] = "UTC"
                stat[var] = vals
                unit = self.UNIT_MAP[units[0]]
                stat["var_info"][var] = dict(units=unit)
                stats.append(stat)
        return stats

    def read_file(self, filename, vars_to_retrieve=None):
        """
        This method is returns just the raw content of a file as a dict

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results

        Raises
        ------
        NotImplementedError
        """

        # unfortunately the files have different encodings, so we have to try them
        # on the entire file first
        ret_data = {}
        encoding = "utf_8"
        try:
            with open(filename, encoding=encoding) as infile:
                linedata = infile.readlines()
        except UnicodeDecodeError:
            encoding = "cp863"
            with open(filename, encoding=encoding) as infile:
                linedata = infile.readlines()
        except Exception:
            logger.info(f"unforeseen encoding in file {filename}! Trying to determine encoding")
            try:
                encoding = self.get_file_encoding(filename)
                logger.info(f"determined encoding: {encoding}")
                with open(filename, encoding=encoding) as infile:
                    linedata = infile.readlines()
            except MemoryError:
                logger.info("could not determine encoding due to MemoryError: Skipping file...")
                raise DataRetrievalError(
                    "could not determine encoding due to MemoryError: Skipping file..."
                )
                return ret_data

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        file_vars_to_retrieve = [self.VAR_MAP[x] for x in vars_to_retrieve]

        tot_lines_retrieved = 0
        for var in vars_to_retrieve:
            ret_data[var] = {}
            ret_data[var]["lines_retrieved"] = 0
            ret_data[var]["linedata"] = []

            for line in linedata:
                line_arr = line.split(self.FILE_COL_DELIM)
                # skip malformed lines
                if len(line_arr) != self.FILE_COL_ROW_NUMBER:
                    continue
                # skip lines that do not contain data of an interesting variable
                if line_arr[self.ROW_VAR_COL] not in file_vars_to_retrieve:
                    continue

                # make the numerical values numerical here already
                line_arr[4] = float(line_arr[4])
                line_arr[7] = float(line_arr[7])
                ret_data[var]["linedata"].append(line_arr)
                ret_data[var]["lines_retrieved"] += 1
                tot_lines_retrieved += 1
        return ret_data

    def read(self, vars_to_retrieve=None, first_file=None, last_file=None):
        """
        Read variable data

        Parameters
        ----------
        vars_to_retrieve : str or list, optional
            List of variables to be retrieved. The default is None.
        first_file : int, optional
            Index of first file to be read. The default is None, in which case
            index 0 in file list is used.
        last_file : int, optional
            Index of last file to be read. The default is None, in which case
            last index in file list is used.

        Returns
        -------
        data : UngriddedData
            loaded data object.

        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        self._init_station_metadata()
        files = self.get_file_list()
        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)
        files = files[first_file:last_file]

        stats = self._read_files(files, vars_to_retrieve)

        data = UngriddedData.from_station_data(
            stats, add_meta_keys=["timezone", "area_classification", "station_classification"]
        )

        return data

    # the following has been stolen from
    # https://stackoverflow.com/questions/44402983/how-to-read-the-file-without-encoding-and-extract-desired-urls-with-python3
    def get_file_bom_encoding(self, filename):
        with open(filename, "rb") as openfileobject:
            line = str(openfileobject.readline())
            if line[2:14] == str(codecs.BOM_UTF8).split("'")[1]:
                return "utf_8"
            if line[2:10] == str(codecs.BOM_UTF16_BE).split("'")[1]:
                return "utf_16"
            if line[2:10] == str(codecs.BOM_UTF16_LE).split("'")[1]:
                return "utf_16"
            if line[2:18] == str(codecs.BOM_UTF32_BE).split("'")[1]:
                return "utf_32"
            if line[2:18] == str(codecs.BOM_UTF32_LE).split("'")[1]:
                return "utf_32"
        return ""

    def get_all_file_encodings(self, filename):
        encoding_list = []
        encodings = (
            "utf_8",
            "cp863",
            "utf_16",
            "utf_16_le",
            "utf_16_be",
            "utf_32",
            "utf_32_be",
            "utf_32_le",
            "cp850",
            "cp437",
            "cp852",
            "cp1252",
            "cp1250",
            "ascii",
            "utf_8_sig",
            "big5",
            "big5hkscs",
            "cp037",
            "cp424",
            "cp500",
            "cp720",
            "cp737",
            "cp775",
            "cp855",
            "cp856",
            "cp857",
            "cp858",
            "cp860",
            "cp861",
            "cp862",
            "cp864",
            "cp865",
            "cp866",
            "cp869",
            "cp874",
            "cp875",
            "cp932",
            "cp949",
            "cp950",
            "cp1006",
            "cp1026",
            "cp1140",
            "cp1251",
            "cp1253",
            "cp1254",
            "cp1255",
            "cp1256",
            "cp1257",
            "cp1258",
            "euc_jp",
            "euc_jis_2004",
            "euc_jisx0213",
            "euc_kr",
            "gb2312",
            "gbk",
            "gb18030",
            "hz",
            "iso2022_jp",
            "iso2022_jp_1",
            "iso2022_jp_2",
            "iso2022_jp_2004",
            "iso2022_jp_3",
            "iso2022_jp_ext",
            "iso2022_kr",
            "latin_1",
            "iso8859_2",
            "iso8859_3",
            "iso8859_4",
            "iso8859_5",
            "iso8859_6",
            "iso8859_7",
            "iso8859_8",
            "iso8859_9",
            "iso8859_10",
            "iso8859_13",
            "iso8859_14",
            "iso8859_15",
            "iso8859_16",
            "johab",
            "koi8_r",
            "koi8_u",
            "mac_cyrillic",
            "mac_greek",
            "mac_iceland",
            "mac_latin2",
            "mac_roman",
            "mac_turkish",
            "ptcp154",
            "shift_jis",
            "shift_jis_2004",
            "shift_jisx0213",
        )
        for e in encodings:
            try:
                fh = codecs.open(filename, "r", encoding=e)
                fh.readlines()
            except UnicodeDecodeError:
                fh.close()
            except UnicodeError:
                fh.close()
            else:
                encoding_list.append([e])
                fh.close()
                continue
        return encoding_list

    def get_file_encoding(self, filename):
        file_encoding = self.get_file_bom_encoding(filename)
        if file_encoding != "":
            return file_encoding
        encoding_list = self.get_all_file_encodings(filename)
        file_encoding = str(encoding_list[0][0])
        if file_encoding[-3:] == "_be" or file_encoding[-3:] == "_le":
            file_encoding = file_encoding[:-3]
        return file_encoding



from __future__ import annotations

from pathlib import Path

from pyaerocom.config import ALL_REGION_NAME

from .common import add_dummy_model_data

START = 2000
STOP = 2020
lat_range = (30, 60)
lon_range = (10, 40)


def fake_model_data(tmp_path: str | Path) -> dict:
    for i, year in enumerate(range(START, STOP), start=1):
        model_data_dir = add_dummy_model_data(
            "prmm",
            "mm",
            "monthly",
            "Surface",
            name="DUMMY-MOD-TRENDS",
            year=year,
            lat_range=lat_range,
            lon_range=lon_range,
            value=i,
            tmp_path=tmp_path,
        )
    return dict(
        DUMMY=dict(
            model_id="DUMMY-MODEL",
            model_data_dir=model_data_dir,
        )
    )


def fake_obs_data(tmp_path: str | Path) -> dict:
    for i, year in enumerate(range(START, STOP), start=1):
        obs_data_dir = add_dummy_model_data(
            "prmm",
            "mm d-1",
            "monthly",
            "Surface",
            name="DUMMY-OBS-TRENDS",
            year=year,
            lat_range=lat_range,
            lon_range=lon_range,
            value=i,
            tmp_path=tmp_path,
        )

    return dict(
        DUMMY=dict(
            obs_id="DUMMY-OBS",
            obs_vars=("prmm",),
            obs_data_dir=obs_data_dir,
            obs_vert_type="Surface",
        )
    )


CFG = dict(
    model_cfg=dir(),  # fake_model_data("PATH_TO_MODEL_DATA"),
    obs_cfg=dir(),  # fake_obs_data("PATH_TO_MODEL_DATA"),
    json_basedir="PATH_TO_AEROVAL_OUT/data",
    coldata_basedir="PATH_TO_AEROVAL_OUT/coldata",
    # if True, existing colocated data files will be deleted
    reanalyse_existing=True,
    raise_exceptions=True,
    only_json=False,
    add_model_maps=False,
    only_model_maps=False,
    clear_existing_json=False,
    # Regional filter for analysis
    filter_name=f"{ALL_REGION_NAME}-wMOUNTAINS",
    # colocation frequency (no statistics in higher resolution can be computed)
    ts_type="monthly",
    map_zoom="World",
    freqs=["monthly"],
    periods=[f"{START}-{STOP}"],
    add_trends=True,
    main_freq="monthly",
    harmonise_units=True,
    proj_id="test",
    exp_id="exp5",
    exp_name="AeroVal test experiment 5",
    exp_descr=("Test setup for more trends evaluation (FAKE PRECIP DATA)"),
    exp_pi="Jonas Gliss",
    public=True,
    # directory where colocated data files are supposed to be stored
    weighted_stats=True,
)



from __future__ import annotations

import os
from collections.abc import Iterable
from pathlib import Path

import pytest

from pyaerocom.aeroval import EvalSetup
from pyaerocom.aeroval._processing_base import DataImporter, HasColocator, HasConfig
from tests.fixtures.aeroval import CAMS84_CONFIG


@pytest.fixture()
def aeroval_path(tmp_path: Path) -> None:
    """temporary path for aeroval outputs"""
    path: Path = tmp_path / "aeroval/aeroval1"
    path.mkdir(parents=True)
    os.chdir(path)


def test_evalsetup_args():
    setup = EvalSetup(proj_id="project", exp_id="experiment")
    assert setup
    assert setup.proj_id == setup.proj_info.proj_id == "project"
    assert setup.exp_id == setup.exp_info.exp_id == "experiment"


def test_evalsetup_kwargs():
    setup = EvalSetup(proj_id="project", exp_id="experiment")
    assert setup
    assert setup.proj_id == setup.proj_info.proj_id == "project"
    assert setup.exp_id == setup.exp_info.exp_id == "experiment"


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(("model_cfg",), id="cams84 models"),
        pytest.param(("obs_cfg",), id="cams84 obs"),
        pytest.param(
            (
                "reanalyse_existing",
                "only_json",
                "add_model_maps",
                "only_model_maps",
                "clear_existing_json",
                "raise_exceptions",
            ),
            id="bool options",
        ),
        pytest.param(
            ("json_basedir", "coldata_basedir", "io_aux_file"),
            id="IO paths",
        ),
        pytest.param(("filter_name",), id="regional filter"),
        pytest.param(
            (
                "ts_type",
                "map_zoom",
                "freqs",
                "periods",
                "main_freq",
                "zeros_to_nan",
                "min_num_obs",
                "colocate_time",
            ),
            id="colocation options",
        ),
        pytest.param(
            ("obs_remove_outliers", "model_remove_outliers"),
            id="outlier options",
        ),
        pytest.param(
            (
                "harmonise_units",
                "regions_how",
                "annual_stats_constrained",
                "weighted_stats",
                "var_order_menu",
            ),
            id="other options",
        ),
        pytest.param(
            ("exp_name", "exp_descr", "exp_pi"),
            id="experiment description",
        ),
    ],
)
def test_evalsetup_cams84(keys: Iterable[str], aeroval_path):
    assert all(k in CAMS84_CONFIG for k in keys)

    config = {k: CAMS84_CONFIG[k] for k in keys}
    assert config

    setup = EvalSetup(
        proj_id=CAMS84_CONFIG["proj_id"],  # type:ignore[arg-type]
        exp_id=CAMS84_CONFIG["exp_id"],  # type:ignore[arg-type]
        **config,
    )
    assert setup


def test_HasConfig():
    setup = EvalSetup(**CAMS84_CONFIG)
    config = HasConfig(setup)
    assert config.raise_exceptions == CAMS84_CONFIG["raise_exceptions"]
    assert config.reanalyse_existing == CAMS84_CONFIG["reanalyse_existing"]


@pytest.mark.parametrize(
    "model",
    [
        None,
        pytest.param(
            "IFS-CTRL",
            marks=[
                pytest.mark.xfail(reason="broken test", raises=KeyError),
                pytest.mark.dependency(name="HasColocator::model"),
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    "obs",
    [
        None,
        pytest.param("EEA-NRT-rural", marks=[pytest.mark.dependency(name="HasColocator::obs")]),
    ],
)
def test_HasColocator(model: str | None, obs: str | None):
    setup = EvalSetup(**CAMS84_CONFIG)  # type:ignore[arg-type]
    config = HasColocator(setup)
    assert config.get_colocator(model_name=model, obs_name=obs)


@pytest.mark.parametrize(
    "model",
    [pytest.param("IFS-CTRL", marks=pytest.mark.dependency(depends="HasColocator::model"))],
)
@pytest.mark.parametrize("var", ["vmro3", "vmrno2"])
def test_DataImporter_read_model_data(model: str, var: str):
    setup = EvalSetup(**CAMS84_CONFIG)  # type:ignore[arg-type]
    assert DataImporter(setup).read_model_data(model, var)


@pytest.mark.parametrize(
    "obs",
    [pytest.param("EEA-NRT-rural", marks=pytest.mark.dependency(depends="HasColocator::obs"))],
)
@pytest.mark.parametrize("var", ["vmro3", "vmrno2"])
def test_DataImporter_read_ungridded_obsdata(obs: str, var: str):
    setup = EvalSetup(**CAMS84_CONFIG)  # type:ignore[arg-type]
    assert DataImporter(setup).read_ungridded_obsdata(obs, var)



from __future__ import annotations

from pathlib import Path

import aerovaldb
import pytest

from pyaerocom import const
from pyaerocom.aeroval import EvalSetup, ExperimentProcessor
from pyaerocom.aeroval.experiment_output import ExperimentOutput, ProjectOutput
from pyaerocom.aeroval.json_utils import read_json, write_json

BASEDIR_DEFAULT = Path(const.OUTPUTDIR) / "aeroval" / "data"


@pytest.fixture()
def dummy_expout(tmp_path: Path) -> ExperimentOutput:
    """ExperimentOutput instance"""
    setup = EvalSetup(proj_id="proj", exp_id="exp", json_basedir=str(tmp_path))
    return ExperimentOutput(setup)


def test_invalid_resource_error():
    with pytest.raises(ValueError):
        ProjectOutput("test", None)


def test_avdb_resource_init(tmp_path):
    with aerovaldb.open(str(tmp_path)) as db:
        val = ProjectOutput("test", db)

        assert val.avdb is db


@pytest.mark.parametrize(
    "proj_id,json_basedir",
    [
        ("bla", "/"),
    ],
)
def test_ProjectOutput(proj_id: str, json_basedir: str):
    val = ProjectOutput(proj_id, json_basedir)
    assert val.proj_id == proj_id
    if json_basedir is not None:
        assert Path(val.json_basedir).exists()


@pytest.mark.parametrize(
    "proj_id,json_basedir,exception,error",
    [
        pytest.param(42, None, ValueError, "need str, got 42", id="ValueError"),
    ],
)
def test_ProjectOutput_error(proj_id, json_basedir, exception: type[Exception], error: str):
    with pytest.raises(exception) as e:
        ProjectOutput(proj_id, json_basedir)
    assert str(e.value) == error


def test_ProjectOutput_proj_dir(tmp_path: Path):
    val = ProjectOutput("test", str(tmp_path))
    path = tmp_path / "test"
    assert val.proj_dir == str(path)
    assert path.exists()


def test_ProjectOutput_experiments_file(tmp_path: Path):
    val = ProjectOutput("test", str(tmp_path))
    path = tmp_path / "test" / "experiments.json"
    assert Path(val.experiments_file) == path


def test_ProjectOutput_available_experiments(tmp_path: Path):
    val = ProjectOutput("test", str(tmp_path))
    assert val.available_experiments == []

    write_json({"exp": 42}, val.experiments_file)
    assert val.available_experiments == ["exp"]


def test_ExperimentOutput__add_entry_experiments_json():
    cfg = EvalSetup(proj_id="proj", exp_id="exp")
    val = ExperimentOutput(cfg)

    val._add_entry_experiments_json("test", 42)
    assert val.available_experiments == ["test"]


def test_ExperimentOutput__del_entry_experiments_json(tmp_path):
    cfg = EvalSetup(proj_id="proj", exp_id="exp")
    val = ExperimentOutput(cfg)

    exp_id = "test"
    assert exp_id in val.available_experiments

    val._del_entry_experiments_json(exp_id)
    assert exp_id not in val.available_experiments


def test_ExperimentOutput():
    cfg = EvalSetup(proj_id="proj", exp_id="exp")
    val = ExperimentOutput(cfg)
    assert isinstance(val.cfg, EvalSetup)
    assert val.proj_id == cfg.proj_info.proj_id

    path = Path(val.json_basedir)
    assert path == BASEDIR_DEFAULT
    assert path.exists()


def test_ExperimentOutput_error():
    with pytest.raises(ValueError):
        ExperimentOutput(None)


def test_ExperimentOutput_exp_id(dummy_expout: ExperimentOutput):
    assert dummy_expout.exp_id == "exp"


def test_ExperimentOutput_exp_dir(dummy_expout: ExperimentOutput, tmp_path: Path):
    assert Path(dummy_expout.exp_dir) == tmp_path / "proj" / "exp"


def test_ExperimentOutput_results_available_False(dummy_expout: ExperimentOutput):
    assert not dummy_expout.results_available


def test_ExperimentOutput_update_menu_EMPTY(dummy_expout: ExperimentOutput):
    dummy_expout.update_menu()

    data = dummy_expout.avdb.get_menu(dummy_expout.proj_id, dummy_expout.exp_id)
    assert data == {}


def test_ExperimentOutput_update_interface_EMPTY(dummy_expout: ExperimentOutput):
    dummy_expout.update_interface()


def test_ExperimentOutput_update_heatmap_json_EMPTY(dummy_expout: ExperimentOutput):
    dummy_expout._sync_heatmaps_with_menu_and_regions()


def test_ExperimentOutput__results_summary_EMPTY(dummy_expout: ExperimentOutput):
    assert dummy_expout._results_summary() == dict(obs=[], ovar=[], vc=[], mod=[], mvar=[], per=[])


def test_ExperimentOutput_clean_json_files_EMPTY(dummy_expout: ExperimentOutput):
    modified = dummy_expout.clean_json_files()
    assert len(modified) == 0


@pytest.mark.parametrize("also_coldata", [True, False])
def test_ExperimentOutput_delete_experiment_data(tmp_path: Path, also_coldata: bool):
    json_path = tmp_path / "json"
    coldata_path = tmp_path / "coldata"
    setup = EvalSetup(
        proj_id="proj",
        exp_id="exp",
        coldata_basedir=coldata_path,
        json_basedir=json_path,
    )
    assert setup.path_manager.coldata_basedir == coldata_path
    assert setup.path_manager.json_basedir == json_path
    assert coldata_path.exists()

    out = ExperimentOutput(setup)
    expdir = json_path / "proj" / "exp"
    coldir = coldata_path / "proj" / "exp"
    assert coldir == Path(out.cfg.path_manager.get_coldata_dir())
    assert expdir == Path(out.exp_dir)
    assert coldir.exists()

    out.delete_experiment_data(also_coldata=also_coldata)
    assert coldata_path.exists()
    assert not expdir.exists()
    assert coldir.exists() == (not also_coldata)


@pytest.mark.parametrize(
    "var,val",
    [
        (
            "ang4487aer",
            {
                "scale": [0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2],
                "colmap": "coolwarm",
            },
        ),
        (
            "concprcpso4",
            {
                "colmap": "coolwarm",
                "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
            },
        ),
    ],
)
def test_ExperimentOutput__get_cmap_info(dummy_expout: ExperimentOutput, var, val):
    assert dummy_expout._get_cmap_info(var) == val


### BELOW ARE TESTS ON ACTUAL OUTPUT THAT DEPEND ON EVALUATION RUNS


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_ExperimentOutput_delete_experiment_data_CFG1(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    cfg.webdisp_opts.regions_how = "htap"
    cfg.webdisp_opts.add_model_maps = False
    cfg.statistics_opts.add_trends = False
    cfg.time_cfg.add_seasons = False
    proc = ExperimentProcessor(cfg)
    proc.run()
    path = Path(proc.exp_output.exp_dir)
    assert path.exists()
    proc.exp_output.delete_experiment_data()
    assert not path.exists()


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_Experiment_Output_clean_json_files_CFG1(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    proc = ExperimentProcessor(cfg)
    proc.run()
    modified = proc.exp_output.clean_json_files()
    assert len(modified) == 0


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_Experiment_Output_clean_json_files_CFG1_INVALIDMOD(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    cfg.model_cfg.add_entry("mod1", cfg.model_cfg.get_entry("TM5-AP3-CTRL"))
    proc = ExperimentProcessor(cfg)
    proc.run()
    cfg.model_cfg.remove_entry("mod1")
    modified = proc.exp_output.clean_json_files()
    assert len(modified) == 13


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_Experiment_Output_clean_json_files_CFG1_INVALIDOBS(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    cfg.obs_cfg.add_entry("obs1", eval_config["obs_cfg"]["AERONET-Sun"])
    proc = ExperimentProcessor(cfg)
    proc.run()
    cfg.obs_cfg.remove_entry("obs1")
    modified = proc.exp_output.clean_json_files()
    assert len(modified) == 13


@pytest.mark.parametrize(
    "add_names,order,result",
    [
        (["c", "b", "a"], None, ["a", "b", "c"]),
        (["c", "b", "a"], ["c", "b", "a"], ["c", "b", "a"]),
        (["c", "b", "a"], [42], ["a", "b", "c"]),
        (["c", "b", "a"], ["b"], ["b", "a", "c"]),
        (["c", "b", "a"], ["b", "c"], ["b", "c", "a"]),
    ],
)
def test_ExperimentOutput_reorder_experiments(
    dummy_expout: ExperimentOutput, add_names, order, result
):
    path = Path(dummy_expout.experiments_file)

    data = dict().fromkeys(add_names, dict(public=True))
    assert list(data) == add_names

    write_json(data, path, indent=4)
    dummy_expout.reorder_experiments(order)
    new = read_json(path)
    assert list(new) == result
    path.unlink()


def test_ExperimentOutput_reorder_experiments_error(dummy_expout: ExperimentOutput):
    with pytest.raises(ValueError) as e:
        dummy_expout.reorder_experiments("b")
    assert str(e.value) == "need list as input"


@pytest.mark.parametrize("cfg,drop_stats,stats_decimals", [("cfgexp1", ("mab", "R_spearman"), 2)])
def test_Experiment_Output_drop_stats_and_decimals(
    eval_config: dict, drop_stats, stats_decimals: int
):
    eval_config["drop_stats"], eval_config["stats_decimals"] = (
        drop_stats,
        stats_decimals,
    )
    cfg = EvalSetup(**eval_config)
    cfg.model_cfg.add_entry("mod1", cfg.model_cfg.get_entry("TM5-AP3-CTRL"))
    proc = ExperimentProcessor(cfg)
    proc.run()
    path = Path(proc.exp_output.exp_dir)
    files = [f for f in path.iterdir() if f.is_file()]
    assert any(["statistics.json" in f.name for f in files])
    statistics_json = read_json(path / "statistics.json")
    assert all([stat not in statistics_json for stat in drop_stats])
    assert all([statistics_json[stat]["decimals"] == stats_decimals for stat in statistics_json])


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_Experiment_Output_stats_reorder(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    requested_keys = ("fge", "R_spatial_mean", "mnmb", "nmb")
    cfg.webdisp_opts.stats_order_menu = requested_keys
    proc = ExperimentProcessor(cfg)
    proc.run()
    path = Path(proc.exp_output.exp_dir)
    statistics_json = read_json(path / "statistics.json")

    retrieved_keys = list(statistics_json.keys())

    assert tuple(retrieved_keys[:4]) == requested_keys



from __future__ import annotations

import numpy as np
import pytest
from iris.coords import DimCoord
from iris.cube import Cube

from pyaerocom.io.aux_read_cubes import (
    add_cubes,
    compute_angstrom_coeff_cubes,
    conc_from_vmr,
    conc_from_vmr_STP,
    divide_cubes,
    mmr_from_vmr,
    mmr_to_vmr_cube,
    multiply_cubes,
    rho_from_ts_ps,
    subtract_cubes,
)


@pytest.fixture
def simple_cube():
    latitude = DimCoord(np.linspace(-90, 90, 4), standard_name="latitude", units="degrees")
    longitude = DimCoord(np.linspace(45, 360, 8), standard_name="longitude", units="degrees")
    cube = Cube(np.ones((4, 8), np.float32), dim_coords_and_dims=[(latitude, 0), (longitude, 1)])
    return cube


def test_add_cubes(simple_cube: Cube):
    cube1, cube2 = simple_cube, simple_cube
    summed_cube = add_cubes(cube1, cube2)
    assert np.all(cube1.data + cube2.data == summed_cube.data)


def test_subtract_cubes(simple_cube: Cube):
    cube1, cube2 = simple_cube, simple_cube
    differenced_cube = subtract_cubes(cube1, cube2)
    assert np.all(cube1.data - cube2.data == differenced_cube.data)


def test_multiple_cubes(simple_cube: Cube):
    cube1, cube2 = simple_cube, simple_cube
    multiplied_cube = multiply_cubes(cube1, cube2)
    assert np.all(cube1.data * cube2.data == multiplied_cube.data)


def test_divide_cubes(simple_cube: Cube):
    cube1, cube2 = simple_cube, simple_cube
    divided_cube = divide_cubes(cube1, cube2)
    assert np.all(cube1.data / cube2.data == divided_cube.data)


def test_compute_angstrom_coeff_cubes(simple_cube: Cube):
    cube1, cube2 = simple_cube, simple_cube
    ang_cube = compute_angstrom_coeff_cubes(cube1, cube2, lambda1=1, lambda2=2)
    assert np.all(ang_cube.data) == 0


def test_rho_from_ts_ps(simple_cube: Cube):
    cube1, cube2 = simple_cube, simple_cube
    rho = rho_from_ts_ps(cube1, cube2)
    assert np.all(rho.data == pytest.approx(287.058, rel=1e-4))


@pytest.mark.parametrize(
    "var_name, expected_result",
    [("air_dry", 1), ("o3", 1.6571896), ("glyox", 2.0036802)],
)
def test_mmr_from_vmr(simple_cube: Cube, var_name: str, expected_result: float):
    cube = simple_cube
    cube.var_name = var_name
    mmr_cube = mmr_from_vmr(cube)
    assert np.all(mmr_cube.data == pytest.approx(expected_result, rel=1e-4))


@pytest.mark.parametrize(
    "var_name, expected_result",
    [("air_dry", 287.058), ("o3", 475.70956), ("glyox", 575.1725)],
)
def test_conc_from_vmr(simple_cube: Cube, var_name: str, expected_result: float):
    cube, ts, ps = simple_cube, simple_cube, simple_cube
    cube.var_name = var_name
    conc_cube = conc_from_vmr(cube, ts, ps)
    assert np.all(conc_cube.data == pytest.approx(expected_result, rel=1e-4))


@pytest.mark.parametrize(
    "var_name, expected_result",
    [("air_dry", 0.8302862), ("o3", 1.3759416), ("glyox", 1.6636281)],
)
def test_conc_from_vmr_STP(simple_cube: Cube, var_name: str, expected_result: float):
    cube = simple_cube
    cube.var_name = var_name
    conc_cube = conc_from_vmr_STP(cube)
    assert np.all(conc_cube.data == pytest.approx(expected_result, rel=1e-4))


@pytest.mark.parametrize(
    "var_name, units, expected_result",
    [
        ("mmro3", "kg kg-1", 6.0343123e08),
        ("mmrso4", "kg kg-1", 3.0152717e08),
    ],
)
def test_mmr_to_vmr_cube(simple_cube: Cube, var_name: str, units: str, expected_result: float):
    cube = simple_cube
    cube.var_name = var_name
    cube.units = units
    vmr_cube = mmr_to_vmr_cube(cube)
    assert np.all(vmr_cube.data == pytest.approx(expected_result, rel=1e-4))



import logging
from traceback import format_exc

import numpy as np
import xarray as xr

from pyaerocom import ColocatedData
from pyaerocom.aeroval._processing_base import HasColocator, ProcessingEngine
from pyaerocom.aeroval.coldatatojson_engine import ColdataToJsonEngine
from pyaerocom.units.datetime import get_lowest_resolution

logger = logging.getLogger(__name__)


class SuperObsEngine(ProcessingEngine, HasColocator):
    """
    Class to handle the processing of combined obs datasets
    """

    def run(self, model_name, obs_name, var_list, try_colocate_if_missing=True):
        self._process_entry(
            model_name=model_name,
            obs_name=obs_name,
            var_list=var_list,
            try_colocate_if_missing=try_colocate_if_missing,
        )

    def _process_entry(self, model_name, obs_name, var_list, try_colocate_if_missing):
        sobs_cfg = self.cfg.obs_cfg.get_entry(obs_name)

        if var_list is None:
            var_list = sobs_cfg.obs_vars
        elif isinstance(var_list, str):
            var_list = [var_list]
        elif not isinstance(var_list, list):
            raise ValueError(f"invalid input for var_list: {var_list}.")

        for var_name in var_list:
            try:
                self._run_var(model_name, obs_name, var_name, try_colocate_if_missing)
            except Exception:
                if self.raise_exceptions:
                    raise
                logger.warning(
                    f"Failed to process superobs entry for {obs_name},  "
                    f"{model_name}, var {var_name}. Reason: {format_exc()}"
                )

    def _run_var(self, model_name, obs_name, var_name, try_colocate_if_missing):
        """
        Run evaluation of superobs entry

        Parameters
        ----------
        model_name : str
            name of model in :attr:`model_config`
        obs_name : str
            name of super observation in :attr:`obs_cfg`
        var_name : str
            name of variable to be processed.
        try_colocate_if_missing : bool
            if True, then missing colocated data objects are computed on the
            fly.

        Raises
        ------
        ValueError
            If multiple (or no) colocated data objects are available for
            individual obs datasets of which the superobservation is comprised.

        Returns
        -------
        None
        """
        coldata_files = []
        coldata_resolutions = []
        vert_codes = []
        obs_entry = self.cfg.obs_cfg.get_entry(obs_name)
        obs_needed = obs_entry.obs_id
        vert_code = obs_entry.obs_vert_type
        for oname in obs_needed:
            fp, ts_type, vert_code = self._get_coldata_fileinfo(
                model_name, oname, var_name, try_colocate_if_missing
            )
            coldata_files.append(fp)
            coldata_resolutions.append(ts_type)
            vert_codes.append(vert_code)

        if len(np.unique(vert_codes)) > 1 or vert_codes[0] != vert_code:
            raise ValueError(
                "Cannot merge observations with different vertical types into super observation..."
            )

        if not len(coldata_files) == len(obs_needed):
            raise ValueError(
                f"Could not retrieve colocated data files for "
                f"all required observations for super obs "
                f"{obs_name}"
            )

        to_freq = get_lowest_resolution(*coldata_resolutions)
        darrs = []
        for fp in coldata_files:
            darrs.append(self._get_dataarray(fp, to_freq, obs_name))

        merged = xr.concat(darrs, dim="station_name")
        coldata = ColocatedData(data=merged)
        engine = ColdataToJsonEngine(self.cfg)
        engine.process_coldata(coldata)

    def _get_dataarray(self, fp, to_freq, obs_name):
        """Get dataarray needed for combination to superobs"""
        data = ColocatedData(data=fp)
        if data.ts_type != to_freq:
            data.resample_time(to_ts_type=to_freq, settings_from_meta=True, inplace=True)
        arr = data.data
        ds = arr["data_source"].values
        source_new = [obs_name, ds[1]]
        arr["data_source"] = source_new  # obs, model_id
        arr.attrs["data_source"] = source_new
        arr.attrs["obs_name"] = obs_name
        return arr

    def _get_coldata_fileinfo(self, model_name, obs_name, var_name, try_colocate_if_missing):
        """Get fileinfo about existing colocated data object"""
        col = self.get_colocator(model_name, obs_name)
        if self.reanalyse_existing:
            col.run(var_list=[var_name])
            cdf = col.files_written
        else:
            cdf = col.get_available_coldata_files([var_name])
            if len(cdf) == 0 and try_colocate_if_missing:
                col.run(var_list=[var_name])
                cdf = col.files_written

        if len(cdf) != 1:
            raise ValueError(
                f"Fatal: Found multiple colocated data objects for "
                f"{model_name}, {obs_name}, {var_name}: {cdf}..."
            )
        fp = cdf[0]
        meta = ColocatedData.get_meta_from_filename(fp)
        ts_type = meta["ts_type"]
        vert_code = self.cfg.obs_cfg.get_entry(obs_name).obs_vert_type
        return (fp, ts_type, vert_code)



from __future__ import annotations

from pathlib import Path

import iris
import numpy as np
from cf_units import Unit

from pyaerocom import GriddedData

# to make things work from iris 3.8
iris.FUTURE.save_split_attrs = True


def make_dummy_cube_3D_daily(
    year=2010, daynum=365, lat_range=None, lon_range=None, value=1, dtype=float
):
    if lat_range is None:
        lat_range = (-30, 30)
    if lon_range is None:
        lon_range = (-10, 10)
    lat_res_deg = lon_res_deg = 5
    times = np.arange(daynum)
    startstr = f"days since {year}-01-01 00:00"
    time_unit = Unit(startstr, calendar="gregorian")

    lons = np.arange(lon_range[0] + lon_res_deg / 2, lon_range[1] + lon_res_deg / 2, lon_res_deg)
    lats = np.arange(lat_range[0] + lat_res_deg / 2, lat_range[1] + lat_res_deg / 2, lat_res_deg)

    latdim = iris.coords.DimCoord(
        lats, var_name="lat", standard_name="latitude", circular=False, units=Unit("degrees")
    )

    londim = iris.coords.DimCoord(
        lons, var_name="lon", standard_name="longitude", circular=False, units=Unit("degrees")
    )

    timedim = iris.coords.DimCoord(times, var_name="time", standard_name="time", units=time_unit)

    latdim.guess_bounds()
    londim.guess_bounds()
    vals = np.ones((len(times), len(lats), len(lons))) * value
    dummy = iris.cube.Cube(vals, units="1")

    dummy.add_dim_coord(latdim, 1)
    dummy.add_dim_coord(londim, 2)
    dummy.add_dim_coord(timedim, 0)
    dummy.var_name = "dummy_grid"

    dummy.data = dummy.data.astype(dtype)
    dummy.attributes["ts_type"] = "daily"
    for coord in dummy.coords():
        coord.points = coord.points.astype(dtype)
    return dummy


def make_griddeddata(var_name, units, ts_type, vert_code, name, **kwargs) -> GriddedData:
    cube = make_dummy_cube_3D_daily(**kwargs)
    cube.var_name = var_name
    cube.units = units

    data = GriddedData(cube)
    data.metadata["data_id"] = name
    data.metadata["vert_code"] = vert_code
    if ts_type != "daily":
        data = data.resample_time(ts_type)
    return data


def add_dummy_model_data(
    var_name, units, ts_type, vert_code, tmp_path: str | Path, name=None, **kwargs
) -> str:
    if name is None:
        name = "DUMMY-MODEL"
    if isinstance(tmp_path, str):
        tmp_path = Path(tmp_path)

    out_path = tmp_path / name / "renamed"
    out_path.mkdir(exist_ok=True, parents=True)
    data = make_griddeddata(var_name, units, ts_type, vert_code, name=name, **kwargs)
    data.to_netcdf(out_dir=out_path)
    return str(out_path)



from pyaerocom.aeroval import EvalSetup
from pyaerocom.aeroval.config.emep.reporting_base import get_CFG


def test_aeroval_config_emep():
    # Setup for models used in analysis
    CFG = get_CFG(
        reportyear=2024,
        year=2021,
        model_dir="/lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.3_metyear2021_emis2022",
    )

    CFG.update(
        dict(
            exp_id="test-2021met_2022emis",
            exp_name="Test runs for 2024 EMEP reporting",
            exp_descr=(
                "Test run from Agnes for 2024_REPORTING/EMEP01_rv5.3_metyear2021_emis2022, i.e. 2021met and 2022emis"
            ),
            exp_pi="S. Tsyro, A. Nyiri, H. Klein",
        )
    )

    # remove "concCocpm10", not in model-output
    for obs in CFG["obs_cfg"]:
        if "concCocpm10" in CFG["obs_cfg"][obs]["obs_vars"]:
            CFG["obs_cfg"][obs]["obs_vars"].remove("concCocpm10")

    # remove "no, pm10, pm25" from EBAS-hourly
    CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concNno")
    CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm10")
    CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm25")

    stp = EvalSetup(**CFG)

    assert stp.periods == ["2021"]
    assert stp.exp_pi == "S. Tsyro, A. Nyiri, H. Klein"
    assert stp.proj_id == "emep"



from __future__ import annotations

import pytest

from pyaerocom.vertical_profile import VerticalProfile


@pytest.fixture(scope="module")
def vertical_profile() -> VerticalProfile:
    return VerticalProfile(
        data=[1, 4, 3, 2, 1],
        altitude=[1, 2, 3, 4, 5],
        dtime=[1],
        var_name="bla",
        data_err=[1, 1, 1, 1, 1],
        var_unit="ug m-3",
        altitude_unit="m",
    )


@pytest.mark.parametrize(
    "kwargs,exception",
    [
        pytest.param(
            dict(),
            TypeError,
            id="no args",
        ),
        pytest.param(
            dict(
                data=[1, 2],
                altitude=[1],
                dtime=[1],
                var_name="bla",
                data_err=[1],
                var_unit="1",
                altitude_unit="1",
            ),
            ValueError,
            id="wrong data",
        ),
    ],
)
def test_VerticalProfile_error(kwargs: dict, exception: type[Exception]):
    with pytest.raises(exception):
        VerticalProfile(**kwargs)



import numpy as np
import pytest

from pyaerocom.exceptions import CoordinateNameError
from pyaerocom.griddeddata import GriddedData
from pyaerocom.vert_coords import (
    AltitudeAccess,
    VerticalCoordinate,
    atmosphere_hybrid_sigma_pressure_coordinate_to_pressure,
    atmosphere_sigma_coordinate_to_pressure,
)


@pytest.fixture
def alt(data_tm5: GriddedData) -> AltitudeAccess:
    return AltitudeAccess(data_tm5)


@pytest.fixture
def sin_wave() -> np.ndarray:
    return np.sin(np.arange(20))


def test_atmosphere_sigma_coordinate_to_pressure_exceptions():
    with pytest.raises(ValueError):
        atmosphere_sigma_coordinate_to_pressure(1.0, 0, "")


@pytest.mark.parametrize(
    "sigma,ps,ptop,expected",
    (
        pytest.param(1.0, 0, 0.0, 0.0, id="ptop-is-float"),
        pytest.param(1.0, 0, np.ones(10), np.zeros(10), id="ptop-is-ndarray"),
        pytest.param(
            0.5,
            0.2,
            np.sin(np.arange(10)),
            np.asarray(
                [
                    0.1,
                    0.52073549,
                    0.55464871,
                    0.17056,
                    -0.27840125,
                    -0.37946214,
                    -0.03970775,
                    0.4284933,
                    0.59467912,
                    0.30605924,
                ]
            ),
            id="sin",
        ),
    ),
)
def test_atmosphere_sigma_coordinate_to_pressure(sigma, ps, ptop, expected):
    result = atmosphere_sigma_coordinate_to_pressure(sigma, ps, ptop)

    assert type(result) is type(ptop)
    if isinstance(result, np.ndarray):
        assert len(result) == len(ptop)

    assert result == pytest.approx(expected)


def test_atmosphere_hybrid_sigma_pressure_coordinate_to_pressure_exceptions():
    with pytest.raises(ValueError):
        atmosphere_hybrid_sigma_pressure_coordinate_to_pressure(np.ones(10), np.ones(5), 1, 0)


@pytest.mark.parametrize(
    "a,b,ps,p0,expected",
    (
        pytest.param(np.ones(10), np.ones(10), 1, None, 2, id="test-p0-is-none"),
        pytest.param(np.ones(10), np.ones(10), 1, 1, 2, id="test-p0-is-not-none"),
        pytest.param(
            np.sin(np.arange(10)),
            np.sin(np.arange(10)),
            0.5,
            0.5,
            np.asarray(
                [
                    0.0,
                    0.84147098,
                    0.90929743,
                    0.14112001,
                    -0.7568025,
                    -0.95892427,
                    -0.2794155,
                    0.6569866,
                    0.98935825,
                    0.41211849,
                ]
            ),
            id="sin",
        ),
    ),
)
def test_atmosphere_hybrid_sigma_pressure_coordinate_to_pressure(
    a: np.ndarray, b: np.ndarray, ps: float, p0: float | None, expected: np.ndarray
):
    result = atmosphere_hybrid_sigma_pressure_coordinate_to_pressure(a, b, ps, p0)

    assert len(a) == len(b) == len(result)
    assert result == pytest.approx(expected)


def test_VerticalCoordinate_exceptions():
    # Unsupported variable name during initialization.
    with pytest.raises(ValueError):
        VerticalCoordinate("jtlkjhsklrg")


def test_VerticalCoordinate_calc_pressure_exceptions():
    # Attempt calculating pressure for unsupported variable.
    vert = VerticalCoordinate("asc")
    with pytest.raises(CoordinateNameError):
        vert.calc_pressure(np.ones(5))


def test_VerticalCoordinate_lev_increases_with_alt_exception():
    vert = VerticalCoordinate("gph")
    with pytest.raises(ValueError, match="Failed to access information"):
        vert.lev_increases_with_alt


@pytest.mark.parametrize(
    "name,expected",
    (
        pytest.param("altitude", True, id="var-name-1"),
        pytest.param("atmosphere_sigma_coordinate", False, id="var-name-2"),
        pytest.param("atmosphere_hybrid_sigma_pressure_coordinate", False, id="var-name-3"),
        pytest.param("asc", False, id="standard-name-1"),
        pytest.param("z", True, id="standard-name-2"),
    ),
)
def test_VerticalCoordinate_lev_increases_with_alt(name: str, expected: bool):
    vert = VerticalCoordinate(name)

    assert vert.lev_increases_with_alt == expected


def test_AltitudeAccess_exceptions():
    with pytest.raises(ValueError):
        AltitudeAccess(None)


def test_AltitudeAccess_search_aux_coords(alt: AltitudeAccess):
    assert alt.search_aux_coords(["lat"])
    assert alt.search_aux_coords("lon")
    assert not alt.search_aux_coords("z")
    assert alt.search_aux_coords(["lat", "lon", "time"])
    assert not alt.search_aux_coords(["lat", "lon", "time", "z"])


def test_AltitudeAccess_search_aux_coords_exception(alt: AltitudeAccess):
    with pytest.raises(CoordinateNameError, match=r"Coordinate .* is not supported"):
        alt.search_aux_coords(["gkjfdshglk"])



from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest

from pyaerocom import VerticalProfile, const
from pyaerocom.io.read_earlinet import ReadEarlinet
from tests.conftest import TEST_RTOL

ROOT: str = const.OBSLOCS_UNGRIDDED["Earlinet-test"]

TEST_FILES: list[str] = [
    f"{ROOT}/EARLINET_AerRemSen_cyc_Lev02_e0355_202104262030_202104262130_v01_qc03.nc",
    f"{ROOT}/EARLINET_AerRemSen_waw_Lev02_b0532_202109221030_202109221130_v01_qc03.nc",
]


def test_all_files_exist():
    for file in TEST_FILES:
        assert Path(file).exists()


@pytest.mark.parametrize(
    "num,vars_to_retrieve",
    [
        (0, "ec355aer"),
    ],
)
def test_ReadEarlinet_read_file(num: int, vars_to_retrieve: list[str]):
    read = ReadEarlinet()
    read.files = paths = TEST_FILES
    stat = read.read_file(paths[num], vars_to_retrieve)

    assert "data_level" in stat
    assert "wavelength_emis" in stat
    assert "has_zdust" in stat
    assert "location" in stat

    if num != 0:
        return

    assert "ec355aer" in stat.var_info
    assert stat.var_info["ec355aer"]["unit_ok"]
    assert stat.var_info["ec355aer"]["err_read"]
    assert stat.var_info["ec355aer"]["outliers_removed"]

    ec355aer = stat.ec355aer
    assert isinstance(ec355aer, VerticalProfile)
    assert len(ec355aer.data) == 164
    assert np.sum(np.isnan(ec355aer.data)) == 0

    assert np.nanmean(ec355aer.data) == pytest.approx(0.02495260001522142, rel=TEST_RTOL)
    assert np.nanstd(ec355aer.data) == pytest.approx(0.03295176956505217, rel=TEST_RTOL)

    assert np.nanmean(ec355aer.data_err) == pytest.approx(0.003919774151078758, rel=TEST_RTOL)
    assert np.nanstd(ec355aer.data_err) == pytest.approx(0.0020847733483625517, rel=TEST_RTOL)

    assert np.min(ec355aer.altitude) == pytest.approx(935.4610692253234, rel=TEST_RTOL)
    assert np.max(ec355aer.altitude) == pytest.approx(10678.245216562595, rel=TEST_RTOL)


@pytest.mark.parametrize(
    "vars_to_retrieve,error",
    [
        ("invalidvar", "invalidvar is not supported"),
        ("od550aer", "od550aer is not supported"),
    ],
)
def test_ReadEarlinet_read_file_error(vars_to_retrieve: str, error: str):
    read = ReadEarlinet()
    read.files = paths = TEST_FILES
    with pytest.raises(ValueError) as e:
        read.read_file(paths[0], vars_to_retrieve)
    assert str(e.value) == error


def test_ReadEarlinet_read():
    read = ReadEarlinet()
    read.files = TEST_FILES
    data = read.read(vars_to_retrieve="ec355aer")

    assert len(data.metadata) == 1
    assert data.shape == (164, 12)

    assert np.nanmin(data._data[:, data._DATAINDEX]) == pytest.approx(
        -0.002188435098876817, rel=TEST_RTOL
    )
    assert np.nanmean(data._data[:, data._DATAINDEX]) == pytest.approx(
        0.02495260001522142, rel=TEST_RTOL
    )
    assert np.nanmax(data._data[:, data._DATAINDEX]) == pytest.approx(
        0.16084047083963124, rel=TEST_RTOL
    )

    merged = data.to_station_data(0)
    # same values as above because only one meta_idx
    assert np.nanmin(merged.ec355aer) == pytest.approx(-0.002188435098876817, rel=TEST_RTOL)
    assert np.nanmean(merged.ec355aer) == pytest.approx(0.02495260001522142, rel=TEST_RTOL)
    assert np.nanmax(merged.ec355aer) == pytest.approx(0.16084047083963124, rel=TEST_RTOL)


@pytest.mark.parametrize(
    "vars_to_retrieve,pattern,num",
    [
        (None, None, 1),
        (["ec355aer"], None, 1),
        (["bsc355aer"], None, 0),
        (["bsc532aer"], None, 1),
    ],
)
def test_ReadEarlinet_get_file_list(
    vars_to_retrieve: list[str] | None, pattern: str | None, num: int
):
    reader = ReadEarlinet("Earlinet-test")
    files = reader.get_file_list(vars_to_retrieve, pattern)
    assert len(files) == num


def test_ReadEarlinet_get_file_list_error():
    reader = ReadEarlinet("Earlinet-test")
    with pytest.raises(NotImplementedError) as e:
        reader.get_file_list(pattern="*e.v*")
    assert str(e.value) == "filetype delimiter . not supported"


def test_ReadEarlinet__get_exclude_filelist():
    reader = ReadEarlinet("Earlinet-test")
    reader.EXCLUDE_CASES.append("onefile.txt")
    files = reader.get_file_list(reader.PROVIDES_VARIABLES)
    assert len(files) == 1
